/**---------------------------------------------------------------------
***  
***  file:    xmeditor.c
***
***  project:  Motif Widgets example programs
***
***  description:  This program demonstrates the Motif text, main window,
***      and dialog widgets, as well as the cut and paste
***      functions.
***  
***  defaults:  xmeditor.c depends on these defaults:
***
#
*allowShellResize:    true
*borderWidth:      0
*highlightThickness:    2
*traversalOn:      true
*keyboardFocusPolicy:    explicit
***-------------------------------------------------------------------*/

/*-------------------------------------------------------------
* information for 'what' 
***-------------------------------------------------------------------*/
char ident1[]="@(#)xmeditor Version A.00.19 17 December 1990";

/*-------------------------------------------------------------
**  Include Files
*/
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <Xm/Xm.h>
#include <Xm/CascadeB.h>
#include <Xm/DialogS.h>
#include <Xm/BulletinB.h>
#include <Xm/FileSB.h>
#include <Xm/MainW.h>
#include <Xm/MessageB.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <Xm/CutPaste.h>

/*-------------------------------------------------------------
**  Global Variables
*/
#define MENU_EXIT    201
#define MENU_OPEN    202
#define MENU_NEW    203
#define MENU_CLOSE    204
#define MENU_SAVE    205
#define MENU_SAVE_AS    206
#define MENU_PRINT    207
#define MENU_CUT    208
#define MENU_COPY    209
#define MENU_COPY_BY_NAME  210
#define MENU_PASTE    211
#define MENU_PASTE_BY_NAME  212
#define MENU_CLEAR    213

#define DIALOG_FSELECT    300
#define DIALOG_CWARNING    301
#define DIALOG_XWARNING    302
#define DIALOG_NEW    303
#define DIALOG_SAVE    304
#define DIALOG_PRINT    305

#define OPEN_HELP    401
#define NEW_HELP    402

Widget text;                 /* Multi-line text widget        */
Widget cut_button;           /* Clipboard cut button         */
Widget copy_button;          /* Clipboard copy button         */
Widget copy_by_name_button;  /* Clipboard copy by name button    */
Widget paste_button;         /* Clipboard paste button    */
Widget paste_by_name_button; /* Clipboard paste by name button    */
Widget clear_button;         /* Clipboard clear button         */
Widget open_dialog;          /* File selection dialog         */
Widget new_dialog;           /* File name prompt dialog         */
Widget close_warning;        /* Special internal selection dialog      */
Widget exit_warning;         /* Special internal selection dialog      */
Widget general_warning;      /* Warning dialog           */
Widget save_dialog;          /* Save as prompt dialog         */
Widget print_warning;        /* Warning dialog           */
Widget CreateVersion ();
void VersionCB ();

Boolean file_saved = True;   /* Indicates that the present file is saved */

char *filename = NULL;       /* String containing file name         */
char *text_buffer;           /* Buffer for text widget text value */
int start_pos, end_pos;      /* Start and end position of last action    */
Boolean by_name = False;     /* Copy by name flag */
int private = 0;  /* private id for clipboard copy by name */

XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
        /* used to set up XmStrings */

char Error[128];

/* Bits for exclamation point in dialog */
char warningBits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xc0, 0x03, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xf0, 0x0f, 0x00, 0x00, 0xf0, 0x0f, 0x00,
   0x00, 0xf0, 0x0f, 0x00, 0x00, 0xf0, 0x0f, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xc0, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00};


/************************************************************************
 *
 *  CreateDefaultImage - create a default image for warning symbol.
 *
 **********************************<->***********************************/
static XImage * CreateDefaultImage (bits, width, height)
char *bits;
int   width, height;
{
    XImage *image;

    image = (XImage *) XtMalloc (sizeof (XImage));
    image->width = width;
    image->height = height;
    image->data = bits;
    image->depth = 1;
    image->xoffset = 0;
    image->format = XYBitmap;
    image->byte_order = LSBFirst;
    image->bitmap_unit = 8;
    image->bitmap_bit_order = LSBFirst;
    image->bitmap_pad = 8;
    image->bytes_per_line = (width+7)/8;
    return (image);
}

/*-------------------------------------------------------------
**  OpenFile
**    Open the present file.  Returns true if file 
**  exists and open is successful.
*/
Boolean OpenFile()
{
   struct stat statbuf;    /* Information on a file. */
   int file_length;    /* Length of file.     */
   unsigned char * file_string;  /* Contents of file.     */
   FILE *fp = NULL;    /* Pointer to open file   */
   
   if ((fp = fopen(filename, "r+")) == NULL)
  if ((fp = fopen(filename, "r")) != NULL) {
      fprintf(stderr, "Warning: file opened read only.\n");
  } else {
      return(False);
  }

   if (stat(filename, &statbuf) == 0)
   file_length = statbuf.st_size;
   else
   file_length = 1000000; /* Arbitrary file length */

   /* Read the file string */
   file_string = (unsigned char *) XtMalloc((unsigned)file_length);
   fread(file_string, sizeof(char), file_length, fp);

   /* Close the file */
   if (fclose(fp) != NULL) fprintf(stderr, "Warning: unable to close file.\n");

   /* Added the file string to the text widget */
   XmTextSetString(text, file_string);

   file_saved = True; /* Intialize to True */
   
   /* Make appropriate item sensitive */
   XtSetSensitive(text, True);
   XtSetSensitive(cut_button, True);
   XtSetSensitive(copy_button, True);
   XtSetSensitive(copy_by_name_button, True);
   XtSetSensitive(clear_button, True);

   return(True);
}

/*-------------------------------------------------------------
**  SaveFile
**    Save the present file.
*/
Boolean SaveFile()
{
    unsigned char * file_string = NULL;     /* Contents of file.          */
    FILE *tfp;           /* Pointer to open temporary file. */
  char  namebuf[BUFSIZ]; /* For "system" call below */
  int    status;
    char *tempname = (char *)XtMalloc(25); /* Temporary file name.         */

    strcpy(tempname, mktemp("/tmp/xmeditXXXXXX"));
    
    if ((tfp = fopen(tempname, "w")) == NULL) {
       fprintf(stderr, "Warning: unable to open temp file, text not saved.\n");
       return(False);;
    }

    /* Get the text string */
    file_string = (unsigned char *)XmTextGetString(text);

    /* Write to a temp file */
    fwrite(file_string, sizeof(char), strlen(file_string) + 1, tfp);

    /* Flush and close the temp file */
    if (fflush(tfp) != NULL) fprintf(stderr,"Warning: unable to flush file.\n");
    if (fclose(tfp) != NULL) fprintf(stderr,"Warning: unable to close file.\n");

    if (file_string != NULL) {
        XmStringFree(file_string); /* free the text string */
    }

    /* Move the tempname to the saved file, but do it independent
     of filesystem boundaries */
  sprintf (namebuf, "cp %s %s\0", tempname, filename);
  status = system(namebuf);
  unlink (tempname);
  if (status == 0) {
        file_saved = True;
    } else {
        fprintf(stderr, "Warning: unable to save file.\n");
        XmStringFree(tempname);
        return(False);
    }
           
    XmStringFree(tempname);
    return(True);
}

/*-------------------------------------------------------------
**      CloseFile
**              Close the present file.
*/
void CloseFile()
{
    /* Zero out the text string in the text widget.
       caution: is causes a value changed callack. */
    XmTextSetString(text, "");

    file_saved = True; /* reinitialize file_saved flag */

    /* Free the file name */
    if (filename != NULL) {
  XmStringFree(filename);
        filename = NULL;
    }

    /* Set text to insensitive */
    XtSetSensitive(text, False);
}

/*-------------------------------------------------------------
**  CopyByName
**    Copy data item to the clipboard by name.
*/
void CopyByName(widget, dataid, privateid, reason)
Widget widget;
unsigned long *dataid;
int *privateid;
int *reason;
{
  int buffer_size, status;

  if (*reason == XmCR_CLIPBOARD_DATA_REQUEST) {
   buffer_size = strlen (text_buffer);
   status = XmClipboardCopyByName (XtDisplay(text), XtWindow(text),
    *dataid, text_buffer, buffer_size, *privateid);
     if (status == ClipboardLocked) {
     printf ("xmeditor procedure CopyByName: Clipboard is locked.");
     return;
     }
  }
  if (*reason == XmCR_CLIPBOARD_DATA_DELETE) 
   printf ("Clipboard data delete request occurred.");
}

/*-------------------------------------------------------------
**  CopyFileToClipboard
**    Copy the present file to the clipboard.
*/
void CopyFileToClipboard(time)
Time time;
{
   char *selected_string = XmTextGetSelection (text); /* Text selection    */
   unsigned long item_id = 0;            /* Clipboard item id */
   int data_id = 0;              /* Clipboard data id */
   int status = 0;              /* Clipboard status  */
   XmString clip_label;
   
   if (by_name) {
   /* Copy the selected by name text to the clipboard */
   by_name = False;
   text_buffer = XmTextGetString (text);  /* Text in the text widget */
   private ++;  /* Unique data name used in call to XmClipboardCopy */
  clip_label = XmStringCreateLtoR ("XM_EDITOR", charset);
  /* Start copy to clipboard, and continue till
     a successful start copy is made */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardStartCopy (XtDisplay(text), XtWindow(text),
                 clip_label, time,
                 text, CopyByName, &item_id);

     /* XmClipboardCopy will put the name of the data on the clipboard */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardCopy (XtDisplay(text), XtWindow(text),
              item_id, "STRING", NULL, 0, private, &data_id);
  /* End the copy to the clipboard and continue till
     a successful end copy is made */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardEndCopy (XtDisplay(text), XtWindow(text),
           item_id);
   
  /* Allow pasting when an item is successfully copied to the clipboard */
  XtSetSensitive(paste_by_name_button, True);

   }
   else {     /* Copy is not by name */
   /* Copy the selected text to the clipboard */
   if (selected_string != NULL) {
  clip_label = XmStringCreateLtoR ("XM_EDITOR", charset);
  /* Start copy to clipboard, and continue until
     a successful start copy is made */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardStartCopy (XtDisplay(text), XtWindow(text),
                 clip_label, time,
                 text, NULL, &item_id);

     /* Move the data to the clipboard, and
     continue until a successful copy is made */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardCopy (XtDisplay(text), XtWindow(text),
              item_id, "STRING", selected_string,
                                      (long)strlen(selected_string)+1, 0,
              &data_id);

  /* End the copy to the clipboard and continue until
     a successful end copy is made */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardEndCopy (XtDisplay(text), XtWindow(text),
           item_id);
   
  /* Allow pasting when an item is successfully copied to the clipboard */
  XtSetSensitive(paste_button, True);

     }
   }
}

/*-------------------------------------------------------------
**  PasteItemFromClipboard
**    Paste item from the clipboard to the current cursor location
*/
void PasteItemFromClipboard()
{
   /* Retrieve the current data from the clipboard
      and paste it at the current cursor position */
   char * selected_string = XmTextGetSelection (text);
                                      /* Contents of selection   */
   int status = 0;                    /* Clipboard status        */
   char *buffer;                      /* Temporary text buffer   */
   int length;                        /* Length of buffer        */
   int outlength = 0;                 /* Length of bytes copie   */
   int private_id = 0;                /* Id of item on clipboard */
   XmTextPosition cursorPos;          /* Text cursor position    */
   register int n;                    /* Arg count               */
   Arg args[10];                      /* Arg list                */

   /* Find the length of the paste item, continue until the length is found */
   while (status != ClipboardSuccess) {
     status = XmClipboardInquireLength(XtDisplay(text), XtWindow(text),
              "STRING", &length);
     if (status == ClipboardNoData) {
  length = 0;
  break;
     }
   }
      
   if (length == 0) {
  fprintf(stderr, "Warning: paste failed, no items to paste.\n");  
  return;
   }

   /* Malloc to necessary space */
   buffer = XtMalloc(length);

   status = XmClipboardRetrieve (XtDisplay(text), XtWindow(text), "STRING",
         buffer, length, &outlength, &private_id);
      
   /* Error in pasting */
   if (status != ClipboardSuccess) {
  fprintf(stderr, "Warning: paste failed, status = %d\n", status);  
  return;
   }

   /* Get cursor position for pasting */
   XtSetArg(args[0], XmNcursorPosition, &cursorPos);
   XtGetValues(text, args, 1);
      
   /* Add new text */
   XmTextReplace(text, cursorPos, cursorPos, buffer);
}

/*-------------------------------------------------------------
**  PasteByName
**    paste by name from the clipboard to the current cursor location
*/
void PasteByName()
{
   int status = 0;           /* Clipboard status        */
   char *buffer;             /* Temporary text buffer   */
   int length;               /* Length of buffer        */
   int outlength = 0;        /* Length of bytes copied  */
   int private_id = 0;       /* Id of item on clipboard */
   XmTextPosition cursorPos; /* Text cursor position    */
   register int n;           /* Arg count               */
   Arg args[10];             /* Arg list                */

   /* find the length of the paste item, continue until the length is found */
   length = strlen(text_buffer);
   if (length == 0) {
  fprintf(stderr, "Warning: paste by name failed, no items to paste.\n");  
  return;
   }

   /* Malloc to necessary space */
   buffer = XtMalloc(length);

   status = XmClipboardRetrieve (XtDisplay(text), XtWindow(text), "STRING",
         buffer, length, &outlength, &private_id);
      
   /* Error in pasting */
   if (status != ClipboardSuccess) {
  fprintf(stderr, "Warning: paste by name failed, status = %d\n", status);  
  return;
   }

   /* Get cursor position for pasting */
   XtSetArg(args[0], XmNcursorPosition, &cursorPos);
   XtGetValues(text, args, 1);
      
   /* Add new text */
   XmTextReplace(text, cursorPos, cursorPos, buffer);
}

/*-------------------------------------------------------------
**  VersionCB      - callback for help button
*/
void VersionCB (w, client_data, call_data) 
Widget    w;    /*  widget id    */
XtPointer    client_data;  /*  data from application   */
XtPointer    call_data;  /*  data from widget class  */
{
  Widget    message_box;    /*  message box    */


  /*  Create message window. */
  message_box = CreateVersion (w);


  /*  Display version window. */
  XtManageChild (message_box);
}

/*-------------------------------------------------------------
**  CreateVersion    - create version window
*/
Widget CreateVersion (parent) 
  Widget    parent;    /*  parent widget  */
{
  Widget    button;
  Widget    message_box;  /*  Message Dialog   */
  Arg    args[10];  /*  arg list    */
  register int  n;    /*  arg count    */

  static char  message[BUFSIZ];  /*  help text  */
  XmString  title_string = NULL;
  XmString  message_string = NULL;
  XmString  button_string = NULL;
  char text4[50];


  strcpy (text4, &ident1[4]);
  /*  Generate message to display. */
  sprintf (message, "This is %s \0",text4);

  /* Create the compound strings */
  message_string = XmStringCreateLtoR (message, charset);
  button_string = XmStringCreateLtoR ("Close", charset);
  title_string = XmStringCreateLtoR ("xmeditor help", charset);


  /*  Create message box dialog. */
  n = 0;
  XtSetArg (args[n], XmNdialogTitle, title_string);  n++;
  XtSetArg (args[n], XmNokLabelString, button_string);  n++;
  XtSetArg (args[n], XmNmessageString, message_string);  n++;
  message_box = XmCreateMessageDialog (parent, "versionbox", args, n);

  button = XmMessageBoxGetChild (message_box, XmDIALOG_CANCEL_BUTTON);
  XtUnmanageChild (button);
  button = XmMessageBoxGetChild (message_box, XmDIALOG_HELP_BUTTON);
  XtUnmanageChild (button);


  /*  Free strings and return message box. */
  if (title_string) XmStringFree (title_string);
  if (message_string) XmStringFree (message_string);
  if (button_string) XmStringFree (button_string);
  return (message_box);
}

/*-------------------------------------------------------------
**  FileChangedCB
**    Process callback from Text.
*/
void FileChangedCB (w, client_data, call_data) 
Widget    w;    /*  widget id    */
XtPointer    client_data;  /*  data from application   */
XtPointer    call_data;  /*  data from widget class  */
{
    /* set the file_saved flag to indicate that the
       file has been modified and the user should be
       notified before exiting. */

    file_saved = False;
}

/*-------------------------------------------------------------
**  MenuCB
**    Process callback from PushButtons in PulldownMenus.
*/
void MenuCB (w, client_data, call_data) 
Widget          w;            /*  Widget id    */
XtPointer       client_data;  /*  Data from application   */
XtPointer       call_data;    /*  Data from widget class  */
{
  register int  n;            /* Arg count        */
  Arg           args[10];     /* Arg list        */
  char          *command;     /* Command used in printing */
  XmString      text_string;  /* Compound string for filename */

  switch ((int)client_data)
  {
    case MENU_OPEN:
      /* Display the file selection dialog */
      XtManageChild (open_dialog);
      break;

    case MENU_NEW:
      /* Display the prompt dialog */
      XtManageChild (new_dialog);
      break;

    case MENU_CLOSE:
       /* The present file has not been saved since
         the last modification */
      if (!file_saved) /* Display the 'save' message dialog */
         XtManageChild (close_warning);
      else
         CloseFile();
      break;

    case MENU_SAVE:
      /* Open a temp file for writing */
      SaveFile();
      break;

    case MENU_SAVE_AS:
      /* Display the 'save as' dialog with the
         present filename displayed in it. */
      text_string = XmStringCreateLtoR (filename, charset);
      n = 0;
      XtSetArg(args[n], XmNtextString,text_string);  n++;
      XtSetValues(save_dialog, args, n);
      XtManageChild (save_dialog);
	  XmStringFree (text_string);
      break;

    case MENU_PRINT:
      if (!file_saved)
         XtManageChild(print_warning);
      else if (filename != NULL) {
        /* Malloc space for the command name. 
              Note: command = size of the filename +
              "lp " + null terminator */
        command = XtMalloc(strlen(filename) + 4);
        sprintf(command, "lp %s\0", filename);
        if (system(command) != NULL)
          fprintf(stderr, "print failed");
        XmStringFree(command);
      }
      break;

    case MENU_EXIT:
      /* Exit if there is no file open */
      if (!file_saved) /* Display the 'save' message dialog */
         XtManageChild (exit_warning);
      else {
         /* Close file pointers and descriptors */
         CloseFile();

         /* Exit this program */
         exit (0);
      }
      break;

    case MENU_CUT:
      {

         /* Call routine to copy selection to clipboard */
         CopyFileToClipboard(((XmPushButtonCallbackStruct *) call_data)->event->xbutton.time);

         /* Call routine to delete primary selection */
         XmTextRemove(text);
          }
      break;

    case MENU_COPY:
          {

         /* Call routine to copy selection to clipboard */
        CopyFileToClipboard(((XmPushButtonCallbackStruct *)call_data)->event->xbutton.time);
          }
      break;

    case MENU_COPY_BY_NAME:
          {

        by_name = True;
         /* Call routine to copy selection to clipboard */
        CopyFileToClipboard(((XmPushButtonCallbackStruct *)call_data)->event->xbutton.time);
          }
      break;

    case MENU_PASTE:
      /* Call the routine that pastes the
         text at the cursor position */
      PasteItemFromClipboard();
      break;

    case MENU_PASTE_BY_NAME:
      /* Call the routine that pastes the by name
         text at the cursor position */
      PasteByName();
      break;

    case MENU_CLEAR:
      /* Call routine to delete primary selection */
      XmTextRemove(text);
      break;

    default:
      /* Unknown client_data was recieved and
         there is no setup to handle this */
      fprintf(stderr, "Warning: in menu callback\n");
      break;
  }

}

/*-------------------------------------------------------------
**  DialogApplyCB
**    Process callback from Dialog apply actions.
*/
static void DialogApplyCB (w, client_data, call_data) 
Widget    w;   /*  Widget id    */
XtPointer    client_data;  /*  Data from application   */
XtPointer    call_data;  /*  Data from widget class  */
{
  char *command;      /* Command used in printing */

  switch ((int)client_data)
  {
    case DIALOG_PRINT:
      if (filename != NULL) {
             /* Malloc space for the command name. 
                Note: command = size of the filename +
                "lp " + null terminator */
        command = XtMalloc(strlen(filename) + 4);
        sprintf(command, "lp %s\0", filename);
        if (system(command) != NULL)
          fprintf(stderr, "print failed");
        XmStringFree(command);
      }
    case DIALOG_CWARNING:
      CloseFile();
      file_saved = True; /* Reset the default */
      break;

    case DIALOG_XWARNING:
      CloseFile();
      exit();
      break;    

    default:
      /* Unknown client_data was received and
         there is no setup to handle this */
      fprintf (stderr, "Warning: in apply callback\n");
      break;

  }
}

/*-------------------------------------------------------------
**  DialogCancelCB
**    Process callback from Dialog cancel actions.
*/
static void DialogCancelCB (w, client_data, call_data) 
Widget    w;    /*  Widget id    */
XtPointer    client_data;  /*  Data from application   */
XtPointer    call_data;  /*  Data from widget class  */
{
  switch ((int)client_data)
  {
    case DIALOG_FSELECT:
      /* Popdown the file selection box */
        XtUnmanageChild (open_dialog);
      break;

    case DIALOG_CWARNING:
    case DIALOG_XWARNING:
    case DIALOG_NEW:
    case DIALOG_PRINT:
    case DIALOG_SAVE:
      /* No action is necessary at this time */
      break;

    default:
      /* An unknown client_data was received and
         there is no setup to handle this */
      fprintf (stderr, "Warning: in cancel callback\n");
      break;
  }
}

/*-------------------------------------------------------------
**  DialogAcceptCB
**    Process callback from Dialog actions.
*/
static void DialogAcceptCB (w, client_data, call_data) 
Widget    w;    /*  Widget id    */
XtPointer    client_data;  /*  Data from application   */
XtPointer    call_data;  /*  Data from widget class  */
{
  char *command;    /* Command used in printing */

  switch ((int)client_data)
  {
    case DIALOG_FSELECT:
            /* Open the file and read it into the text widget */
      if (filename != NULL) {
         XmStringFree(filename);
         filename = NULL;
                        }
      {
         XmFileSelectionBoxCallbackStruct *fcb =
         (XmFileSelectionBoxCallbackStruct *) call_data;

         /* Get the filename from the file selection box */
         XmStringGetLtoR(fcb->value, charset, &filename);

         /* Open file, print error if it does not exist. */
         if (!OpenFile())
               fprintf(stderr, "Warning: unable to open file\n");

         /* Popdown the file selection box */
           XtUnmanageChild (open_dialog);
      }
      break;

    case DIALOG_NEW:
            /* Open the file and read it into the text widget */
      if (filename != NULL) {
         XmStringFree(filename);
         filename = NULL;
                        }
      {
         XmSelectionBoxCallbackStruct *scb =
         (XmSelectionBoxCallbackStruct *) call_data;

         /* Get the filename string from the file
            name prompt box */
         XmStringGetLtoR(scb->value, charset, &filename);

         /* Open file if it exists,
            if not set items sensitive */
         if (!OpenFile(True)) {
              /* make appropriate item sensitive */
              XtSetSensitive(text, True);
              XtSetSensitive(cut_button, True);
              XtSetSensitive(copy_button, True);
              XtSetSensitive(copy_by_name_button, True);
              XtSetSensitive(clear_button, True);
         }

         /* Popdown the file selection box */
           XtUnmanageChild (new_dialog);
      }
      break;

    case DIALOG_CWARNING:
      /* Save the file */
      if (SaveFile()) {
         CloseFile(); /* Close the file */
      } else
         fprintf(stderr, 
             "Warning: unable to save file, file not closed");
      break;

    case DIALOG_XWARNING:
      /* Save the file */
      if (SaveFile()) {
         CloseFile(); /* Close the file */
         exit(0);
                        } else
         fprintf(stderr,
         "Warning: unable to save file, exit aborted");
      break;

    case DIALOG_SAVE:
      {
         XmSelectionBoxCallbackStruct *scb =
         (XmSelectionBoxCallbackStruct *) call_data;

         /* get the filename string from the file
            selection box */
         XmStringGetLtoR(scb->value, charset, &filename);
      
         SaveFile();

         XtUnmanageChild (save_dialog);
      }
      break;

    case DIALOG_PRINT:
      /* Save the file */
      if (SaveFile()) {
         if (filename != NULL) {
             /* Malloc space for the command name. 
                Note: command = size of the filename +
                "lp " + null terminator */
        command = XtMalloc(strlen(filename) + 4);
        sprintf(command, "lp %s\0", filename);
        if (system(command) != NULL)
          fprintf(stderr, "print failed");
        XmStringFree(command);
        }
      } else
         fprintf(stderr, 
            "Warning: unable to save file, file not printed");
      break;

    default:
      /* Unknown callback type */
      fprintf (stderr, "Warning: in accept callback\n");
      break;
  }
}

/*-------------------------------------------------------------
**  CreateSpecialWarningDialog
**    Create special 4 button message box out of a
**  Selection box.
*/
static Widget CreateSpecialWarningDialog (parent, name, image_string, message,
            arglist, argcount)
Widget    parent;
String    name;
String    image_string;
String    message;
ArgList    arglist;
int    argcount;
{
  Widget        warning_dialog; /*  Special warning selection box */
  Widget        work_area;      /*  Rowcolumn for pixmap and text */
  Widget        pixmap_label;   /*  Pixmap label                  */
  Widget        text_label;     /*  Text label                    */
  Widget        apply_button;   /*  Apply button                  */
  Widget        ok_button;      /*  Ok button                     */
  Widget        kid[5];         /*  Buttons                       */
  Pixel         foreground;     /*  Dialog foreground             */
  Pixel         background;     /*  Dialog background             */
  Pixmap        pixmap;         /*  Dialog pixmap                 */
  register int  i;              /*  Kid index                     */
  Arg           args[10];       /*  Arg list                      */
  register int  n;              /*  Arg count                     */


  warning_dialog = XmCreatePromptDialog(parent, name, arglist, argcount);

  n = 0;
  XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
  work_area = XmCreateRowColumn(warning_dialog, "workarea", args, n);
  XtManageChild(work_area);

  n = 0;
  XtSetArg(args[n], XmNforeground, &foreground); n++;
  XtSetArg(args[n], XmNbackground, &background); n++;
  XtGetValues(warning_dialog, args, n);

  n = 0;
  XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
  pixmap = XmGetPixmap(XtScreen(warning_dialog), image_string,
           foreground, background);
  XtSetArg(args[n], XmNlabelPixmap, pixmap); n++;
  pixmap_label = XmCreateLabel(work_area, "pixmap_label", args, n);
  XtManageChild(pixmap_label);

  n = 0;
  XtSetArg(args[n], XmNlabelString,
     XmStringCreateLtoR(message, charset)); n++;
  text_label = XmCreateLabel(work_area, "text_label", args, n);
  XtManageChild(text_label);

        apply_button = XmSelectionBoxGetChild (warning_dialog,
               XmDIALOG_APPLY_BUTTON);
  
  n = 0;
  XtSetArg(args[n], XmNlabelString,
     XmStringCreateLtoR("Discard", charset)); n++;
  XtSetValues(apply_button, args, n);
  XtManageChild(apply_button);

        ok_button = XmSelectionBoxGetChild (warning_dialog,
               XmDIALOG_OK_BUTTON);
  n = 0;
  XtSetArg(args[n], XmNlabelString,
     XmStringCreateLtoR("Save", charset)); n++;
  XtSetValues(ok_button, args, n);

  
        /*      Unmanage unneeded children.
        */
        i = 0;
        kid[i++] = XmSelectionBoxGetChild (warning_dialog, XmDIALOG_TEXT);
        kid[i++] = XmSelectionBoxGetChild (warning_dialog,
                 XmDIALOG_SELECTION_LABEL);
        XtUnmanageChildren (kid, i);

  return(warning_dialog);
}

/*-------------------------------------------------------------
**  DialogHelpCB    - create dialog help window
*/
void DialogHelpCB (w, client_data, call_data)
Widget    w;    /*  Widget id    */
XtPointer    client_data;  /*  Data from application   */
XtPointer    call_data;  /*  Data from widget class  */
{
  Widget    button;
  Widget    message_box;  /*  Message Dialog   */
  Arg    args[10];  /*  Arg list    */
  register int  n;    /*  Arg count    */

  static char  newdlog_help[BUFSIZ];  /*  Help text for new dialog */
  static char  opendlog_help[BUFSIZ];  /*  Help text for open dialog */
  XmString  title_string = NULL;
  XmString  message_string = NULL;
  XmString  button_string = NULL;


  if ((int)client_data == (int)NEW_HELP) {
  /*  Generate help message for new dialog */
  sprintf (newdlog_help, "\
You must move the pointer into the text window and click BSelect \n\
before you enter the new file name.  After you enter the file \n\
name, press Return or move the pointer to 'OK' and click BSelect.  \0");

  message_string = XmStringCreateLtoR (newdlog_help, charset);
  }
  else {
  /*  Generate help message for open dialog */
  sprintf (opendlog_help, "\
Select a file to edit by moving the pointer into the 'Files' box and \n\
over the name of the file you want.  Click BSelect and then move the \n\
pointer to 'OK' and click BSelect.  The text of the file you selected \n\
will appear in the text edit window. \n\
\n\
To change the directory, move the pointer into the 'Directories' box and \n\
over the name of the directory you want.  Click BSelect and then move the \n\
pointer to 'Filter' and click BSelect.  You can also enter a new directory \n\
in the 'File Filter' box by moving the pointer into that box and deleting \n\
any or all of the text therein and entering the desired directory name. \n\
Then, either press Return or move the pointer to 'Filter' and click \n\
BSelect.  You can then select a file as described above. \0");

  message_string = XmStringCreateLtoR (opendlog_help, charset);
  }

  button_string = XmStringCreateLtoR ("Close", charset);
  title_string = XmStringCreateLtoR ("dialog help", charset);


  /*  Create message box dialog.
  */
  n = 0;
  XtSetArg (args[n], XmNdialogTitle, title_string);  n++;
  XtSetArg (args[n], XmNokLabelString, button_string);  n++;
  XtSetArg (args[n], XmNmessageString, message_string);  n++;
  message_box = XmCreateMessageDialog (w, "helpbox", args, n);
  XtManageChild (message_box);

  button = XmMessageBoxGetChild (message_box, XmDIALOG_CANCEL_BUTTON);
  XtUnmanageChild (button);
  button = XmMessageBoxGetChild (message_box, XmDIALOG_HELP_BUTTON);
  XtUnmanageChild (button);


  /*  Free strings and return message box.
  */
  if (title_string) XmStringFree (title_string);
  if (message_string) XmStringFree (message_string);
  if (button_string) XmStringFree (button_string);
}

/*-------------------------------------------------------------
**  CreateHelp    - create help window
*/
Widget CreateHelp (parent) 
  Widget    parent;    /*  Parent widget  */
{
  Widget    button;
  Widget    message_box;  /*  Message Dialog   */
  Arg    args[10];  /*  Arg list    */
  register int  n;    /*  Arg count    */

  static char  message[BUFSIZ];  /*  Help text  */
  XmString  title_string = NULL;
  XmString  message_string = NULL;
  XmString  button_string = NULL;



  /*  Generate message to display.
  */
  sprintf (message, "\
Before you can enter text you must open an existing file or create a new \n\
file.  Move the pointer to 'File' and click BSelect, then select 'Open' \n\
or 'New'.  In either case, dialogs will appear that will allow you to \n\
choose an existing file or enter the name of a new file.  If you choose \n\
a new file, you must move the pointer into the text window and click \n\
BSelect before you enter the new file name.  After you enter the file \n\
name, press Return or move the pointer to 'OK' and click BSelect.  \n\
\n\
The 'Edit' menu has 'Cut', 'Copy', 'Paste', and 'Clear' buttons.  You \n\
must first make the text to be acted upon the primary selection.  The \n\
'Cut' button deletes or cuts the selected text, but saves it on the \n\
clipboard so you can paste it elsewhere.  'Copy' copies the selected \n\
text to the clipboard but does not delete it.  'Paste' pastes the text \n\
currently in the clipboard to the location of the insertion cursor.  \n\
'Clear' deletes the selected text but does not copy it to the clipboard. \0");

  message_string = XmStringCreateLtoR (message, charset);
  button_string = XmStringCreateLtoR ("Close", charset);
  title_string = XmStringCreateLtoR ("xmeditor help", charset);


  /*  Create message box dialog.
  */
  n = 0;
  XtSetArg (args[n], XmNdialogTitle, title_string);  n++;
  XtSetArg (args[n], XmNokLabelString, button_string);  n++;
  XtSetArg (args[n], XmNmessageString, message_string);  n++;
  message_box = XmCreateMessageDialog (parent, "helpbox", args, n);

  button = XmMessageBoxGetChild (message_box, XmDIALOG_CANCEL_BUTTON);
  XtUnmanageChild (button);
  button = XmMessageBoxGetChild (message_box, XmDIALOG_HELP_BUTTON);
  XtUnmanageChild (button);


  /*  Free strings and return message box.
  */
  if (title_string) XmStringFree (title_string);
  if (message_string) XmStringFree (message_string);
  if (button_string) XmStringFree (button_string);
  return (message_box);
}

/*-------------------------------------------------------------
**  HelpCB      - callback for help button
*/
void HelpCB (w, client_data, call_data) 
Widget    w;               /*  Widget id              */
XtPointer    client_data;  /*  Data from application  */
XtPointer    call_data;    /*  Data from widget class */
{
  Widget    message_box;   /*  Message box            */


  /*  Create help window.
  */
  message_box = CreateHelp (w);


  /*  Display help window.
  */
  XtManageChild (message_box);
}


/*-------------------------------------------------------------
**  CreateMenuBar
**    Create menu bar in main window
*/
static Widget CreateMenuBar (parent)
Widget    parent;
{
  Widget        menu_bar;
  Widget        cascade;      /*  CascadeButton             */
  Widget        menu_pane;    /*  RowColumn                 */
  Widget        button;       /*  PushButton                */
  XImage        *image;       /*  Image for warning pixmap  */
  Arg           args[10];     /*  Arg list                  */
  register int  n;            /*  Arg count                 */
  XmString      label_string;
  XmString      acc_string;

  /*  Create MenuArea.
  */
  n = 0;
  menu_bar = XmCreateMenuBar (parent, "menu_bar", args, n);

  /*  Create "Options" pulldown menu.
  */
  n = 0;
  menu_pane = XmCreatePulldownMenu (menu_bar, "menu_pane", args, n);

  image = CreateDefaultImage (warningBits, 32, 32);
  XmInstallImage (image, "warning_image");
          
  label_string = XmStringCreateLtoR ("Open", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'O'); n++;
  button = XmCreatePushButton (menu_pane, "Open", args, n);
  XtAddCallback (button, XmNactivateCallback, MenuCB, MENU_OPEN);
  XtManageChild (button);
  XmStringFree (label_string);

  open_dialog = XmCreateFileSelectionDialog(menu_pane,
         "file selection dialog", NULL, 0);

  XtAddCallback (open_dialog, XmNokCallback,
      DialogAcceptCB, DIALOG_FSELECT);
  XtAddCallback (open_dialog, XmNcancelCallback,
      DialogCancelCB, DIALOG_FSELECT);
  XtAddCallback (open_dialog, XmNhelpCallback,
      DialogHelpCB, OPEN_HELP);
  
  label_string = XmStringCreateLtoR ("New", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'N'); n++;
  button = XmCreatePushButton (menu_pane, "New", args, n);
  XtAddCallback (button, XmNactivateCallback, MenuCB, MENU_NEW);
  XtManageChild (button);
  XmStringFree (label_string);

  label_string = XmStringCreateLtoR ("Enter name of new file.", charset);
  n = 0;
  XtSetArg(args[n], XmNselectionLabelString, label_string); n++;
  new_dialog = XmCreatePromptDialog(menu_pane,
         "new file dialog", args, n);
  XtAddCallback (new_dialog, XmNokCallback,
      DialogAcceptCB, DIALOG_NEW);
  XtAddCallback (new_dialog, XmNcancelCallback,
      DialogCancelCB, DIALOG_NEW);
  XtAddCallback (new_dialog, XmNhelpCallback,
      DialogHelpCB, NEW_HELP);
  XmStringFree (label_string);

  label_string = XmStringCreateLtoR ("Close", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'C'); n++;
  button = XmCreatePushButton (menu_pane, "Close", args, n);
  XtAddCallback (button, XmNactivateCallback, MenuCB, MENU_CLOSE);
  XtManageChild (button);
  XmStringFree (label_string);

  close_warning = CreateSpecialWarningDialog(menu_pane, "save_warning",
              "warning_image", "Save Changes?", args, n);

  XtAddCallback (close_warning, XmNapplyCallback,
           DialogApplyCB, DIALOG_CWARNING);
  XtAddCallback (close_warning, XmNokCallback,
           DialogAcceptCB, DIALOG_CWARNING);

  label_string = XmStringCreateLtoR ("Save", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'S'); n++;
  button = XmCreatePushButton (menu_pane, "Save", args, n);
  XtAddCallback (button, XmNactivateCallback, MenuCB, MENU_SAVE);
  XtManageChild (button);
  XmStringFree (label_string);

  label_string = XmStringCreateLtoR ("Save As..", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'A'); n++;
  button = XmCreatePushButton (menu_pane, "Save As...", args, n);
  XtAddCallback (button, XmNactivateCallback, MenuCB, MENU_SAVE_AS);
  XtManageChild (button);
  XmStringFree (label_string);

  label_string = XmStringCreateLtoR ("Save As..", charset);
  n = 0;
  XtSetArg(args[n], XmNselectionLabelString, label_string); n++;
  save_dialog = XmCreatePromptDialog(menu_pane, "save dialog", args, n);
  XtAddCallback (save_dialog, XmNokCallback,
           DialogAcceptCB, DIALOG_SAVE);
  XmStringFree (label_string);

  label_string = XmStringCreateLtoR ("Print", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'P'); n++;
  button = XmCreatePushButton (menu_pane, "Print", args, n);
  XtAddCallback (button, XmNactivateCallback, MenuCB, MENU_PRINT);
  XtManageChild (button);
  XmStringFree (label_string);

  label_string = XmStringCreateLtoR ("Save file before printing?", charset);
  n = 0;
  XtSetArg(args[n], XmNselectionLabelString, label_string); n++;
  print_warning = CreateSpecialWarningDialog(menu_pane, "print_warning",
       "warning_image", "Save file before printing?", args, n);
  XtAddCallback (print_warning, XmNokCallback,
           DialogAcceptCB, DIALOG_PRINT);
  XmStringFree (label_string);

  label_string = XmStringCreateLtoR ("Exit", charset);
  acc_string = XmStringCreateLtoR ("F3", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'E'); n++;
  XtSetArg(args[n], XmNacceleratorText, acc_string); n++;
  XtSetArg(args[n], XmNaccelerator, "<Key>F3:"); n++;
  button = XmCreatePushButton (menu_pane, "Exit", args, n);
  XtAddCallback (button, XmNactivateCallback, MenuCB, MENU_EXIT);
  XtManageChild (button);
  XmStringFree (label_string);
  XmStringFree (acc_string);

  exit_warning = CreateSpecialWarningDialog(menu_pane, "exit warning",
              "warning_image", "Save Changes?", args, n);
  XtAddCallback (exit_warning, XmNapplyCallback,
           DialogApplyCB, DIALOG_XWARNING);
  XtAddCallback (exit_warning, XmNokCallback,
      DialogAcceptCB, DIALOG_XWARNING);

  label_string = XmStringCreateLtoR ("File", charset);
  n = 0;
  XtSetArg (args[n], XmNsubMenuId, menu_pane);  n++;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'F'); n++;
  cascade = XmCreateCascadeButton (menu_bar, "File", args, n);
  XtManageChild (cascade);
  XmStringFree (label_string);

  /*  Create "Options" pulldown   menu.
  */
  n = 0;
  menu_pane = XmCreatePulldownMenu (menu_bar, "menu_pane", args, n);

  label_string = XmStringCreateLtoR ("Cut", charset);
  acc_string = XmStringCreateLtoR ("Shift+Del", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 't'); n++;
  XtSetArg(args[n], XmNacceleratorText, acc_string); n++;
  XtSetArg(args[n], XmNaccelerator, "Shift<Key>DeleteChar:"); n++;
  cut_button = XmCreatePushButton (menu_pane, "Cut", args, n);
  XtAddCallback (cut_button, XmNactivateCallback, MenuCB, MENU_CUT);
  XtManageChild (cut_button);
  XtSetSensitive(cut_button, False);
  XmStringFree (label_string);
  XmStringFree (acc_string);

  label_string = XmStringCreateLtoR ("Copy", charset);
  acc_string = XmStringCreateLtoR ("Ctrl+Ins", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'C'); n++;
  XtSetArg(args[n], XmNacceleratorText, acc_string); n++;
  XtSetArg(args[n], XmNaccelerator, "Ctrl<Key>InsertChar:"); n++;
  copy_button = XmCreatePushButton (menu_pane, "Copy", args, n);
  XtAddCallback (copy_button, XmNactivateCallback, MenuCB, MENU_COPY);
  XtManageChild (copy_button);
  XtSetSensitive(copy_button, False);
  XmStringFree (label_string);
  XmStringFree (acc_string);

  label_string = XmStringCreateLtoR ("Copy By Name", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'o'); n++;
  copy_by_name_button = XmCreatePushButton (menu_pane, "Copy_By_Name", args, n);
  XtAddCallback (copy_by_name_button, XmNactivateCallback, MenuCB, MENU_COPY_BY_NAME);
  XtManageChild (copy_by_name_button);
  XtSetSensitive(copy_by_name_button, False);
  XmStringFree (label_string);

  label_string = XmStringCreateLtoR ("Paste", charset);
  acc_string = XmStringCreateLtoR ("Shift+Ins", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'P'); n++;
  XtSetArg(args[n], XmNacceleratorText, acc_string); n++;
  XtSetArg(args[n], XmNaccelerator, "Shift<Key>InsertChar:"); n++;
  paste_button = XmCreatePushButton (menu_pane, "Paste", args, n); n++;
  XtAddCallback (paste_button, XmNactivateCallback, MenuCB, MENU_PASTE);
  XtManageChild (paste_button);
  XtSetSensitive(paste_button, False);
  XmStringFree (label_string);
  XmStringFree (acc_string);

  label_string = XmStringCreateLtoR ("Paste By Name", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'a'); n++;
  paste_by_name_button = XmCreatePushButton (menu_pane, "PasteByName", args, n); n++;
  XtAddCallback (paste_by_name_button, XmNactivateCallback, MenuCB, MENU_PASTE_BY_NAME);
  XtManageChild (paste_by_name_button);
  XtSetSensitive(paste_by_name_button, False);
  XmStringFree (label_string);

  label_string = XmStringCreateLtoR ("Clear", charset);
  acc_string = XmStringCreateLtoR ("Del", charset);
  n = 0;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'e'); n++;
  /* Pseudo accelerator - Text already handles
     this action appropriately */
  XtSetArg(args[n], XmNacceleratorText, acc_string); n++;
  clear_button = XmCreatePushButton (menu_pane, "Clear", args, n);
  XtAddCallback (clear_button, XmNactivateCallback, MenuCB, MENU_CLEAR);
  XtManageChild (clear_button);
  XtSetSensitive(clear_button, False);
  XmStringFree (label_string);
  XmStringFree (acc_string);

  label_string = XmStringCreateLtoR ("Edit", charset);
  n = 0;
  XtSetArg (args[n], XmNsubMenuId, menu_pane);  n++;
  XtSetArg(args[n], XmNlabelString, label_string); n++;
  XtSetArg(args[n], XmNmnemonic, 'E'); n++;
  cascade= XmCreateCascadeButton (menu_bar, "Edit", args, n);
  XtManageChild (cascade);
  XmStringFree (label_string);


  /*  Create "Help" pulldown menu. */
  n = 0;
  menu_pane = XmCreatePulldownMenu (menu_bar, "menu_pane", args, n);

  n = 0;
  button = XmCreatePushButton (menu_pane, "Help", args, n);
  XtManageChild (button);
  XtAddCallback (button, XmNactivateCallback, HelpCB, NULL);

  n = 0;
  button = XmCreatePushButton (menu_pane, "Program Version", args, n);
  XtManageChild (button);
  XtAddCallback (button, XmNactivateCallback, VersionCB, NULL);
  
  n = 0;
  XtSetArg (args[n], XmNsubMenuId, menu_pane);  n++;
  cascade = XmCreateCascadeButton (menu_bar, "Help", args, n);
  XtManageChild (cascade);

  n = 0;
  XtSetArg (args[n], XmNmenuHelpWidget, cascade);  n++;
  XtSetValues (menu_bar, args, n);

  return (menu_bar);
}

/*-------------------------------------------------------------
**  CreateText
**    Create Text.
*/
static void CreateText (parent)
Widget    parent;
{
  Arg    args[10];    /*  Arg list    */
  register int  n;    /*  Arg count    */
  register int  i;    /*  Counter    */
     XFontStruct * newfont;

  /* Create text widget */
  n = 0;
  XtSetArg (args[n], XmNrows, 24);  n++;
  XtSetArg (args[n], XmNcolumns, 80);  n++;
  XtSetArg (args[n], XmNresizeWidth, False);  n++;
  XtSetArg (args[n], XmNresizeHeight, False);  n++;
  XtSetArg (args[n], XmNscrollVertical, True);  n++;
  XtSetArg (args[n], XmNscrollHorizontal, True);  n++;
  XtSetArg (args[n], XmNeditMode, XmMULTI_LINE_EDIT);  n++;

  text = XmCreateScrolledText (parent, "text", args, n);

  /* Add value changed callback */
  XtAddCallback (text, XmNmodifyVerifyCallback, FileChangedCB, NULL);

  return;
}

/*-------------------------------------------------------------
**  main
**    Initialize toolkit.
**    Create main window and subwidgets.
**    Realize toplevel widgets.
**    Process events.
*/
void main (argc,argv)
unsigned int  argc;
char     **argv;
{
  Display    *display;        /*  Display             */
  Widget    app_shell;        /*  Application shell   */
  Widget    main;             /*  main window         */
  Widget    menu_bar;         /*  Row column          */
  Widget    form;             /*  Form                */
  XtAppContext  app_context;  /*  Application context */

  Arg    args [10];           /*  Arg list            */
  register int  n;            /*  Arg count           */
  register int  i;            /*  Counter             */

  /*  Initialize toolkit and open display.
  */
  app_shell = XtAppInitialize (&app_context, "XMeditor", NULL, 0, &argc,
  argv, NULL, args, 0);

  /*  Create main window.
  */
  n = 0;
  XtSetArg (args[n], XmNshadowThickness, 0);  n++;
  main = XmCreateMainWindow (app_shell, "main", args, n);
  XtManageChild (main);

  /*  Create menu bar in main window.
  */
  menu_bar = CreateMenuBar (main);
  XtManageChild (menu_bar);

  /*  Create Text.
  */
  CreateText (main);
  XtManageChild (text);

  XmAddTabGroup(text);

  XtSetSensitive(text, False);

  /*  Realize toplevel widgets.
  */
  XtRealizeWidget (app_shell);

  /*  Process events.
  */
  XtAppMainLoop (app_context);
}

